/************************************************************************
   Author: Minh B. Do - Arizona State University
*************************************************************************/
package myplanner.basic_ds;

import java.util.*;

/**
 *  GAction: Represent the grounded action with preconditions are
 *  grounded predicates (Integer) and grounded "Test" (math comparisons). Effects
 *  are also grounded predicates and grounded "Set" (math assignments).
 */
public class GAction {
    int id = -1; // Unique ID for a ground action

    boolean d_type; // The duration is CONSTANT (T), or dynamic (F)
    float d_static; // Value of duration if CONSTANT
    GMathForm d_dynamic;

    boolean c_type; // The cost is CONSTANT (T), or dynamic (F)
    float c_static; // Value of cost if CONSTANT
    GMathForm c_dynamic;

    int[] preconds; // Predicate list represent preconditions
    GTest[] testList; // preconds for metric source (list of functions)

    int[] deleteList, addList;
    GMySet[] setList; // effects for metric resource

    int[] preTime, testTime, addTime, deleteTime, setTime; // Time (start/end) each
    // precondition, test, add or delete occurs. Can be changed to "boolean" now.
    int nPre, nTest, nAdd, nDel, nSet;

    public GAction() {

    }

    /** Constructor */
    public GAction(int numPrecond, int numTest, int numDel, int numAdd, int numSet) {
	nPre = numPrecond;
	nTest = numTest;
	nAdd = numAdd;
	nDel = numDel;
	nSet = numSet;

	d_type = true;
	d_static = 0;

	c_type = true;
	c_static = 1;

	preconds = new int[numPrecond];
	preTime = new int[numPrecond];

	testList = new GTest[numTest];
	testTime = new int[numTest];

	addList = new int[numAdd];
	addTime = new int[numAdd];

	deleteList = new int[numDel];
	setList = new GMySet[numSet];

	deleteTime = new int[numDel];
	setTime = new int[numSet];
    }

    /** Unique ID of this ground action. Which is also the index in the
     *  ground action list */
    public void setID(int i) {
	id = i;
    }

    public int getID() {
	return id;
    }

    /** Duration type of this action. Whether it's fixed (static) duration
     * or a dynamic one (represented by a GMathForm data structure) */
    public void setDType(boolean t) {
	d_type = t;
    }

    public boolean getDType() {
	return d_type;
    }

    public void setDStatic(float f) {
	d_static = f;
    }

    public float getDStatic() {
	return d_static;
    }

    public void setDDynamic(GMathForm m) {
	d_dynamic = new GMathForm(m);
    }

    public GMathForm getDDynamic() {
	return new GMathForm(d_dynamic);
    }

    /** The cost of this action. Represented similar to the duration */
    public void setCType(boolean t) {
	c_type = t;
    }

    public boolean getCType() {
	return c_type;
    }

    public void setCStatic(float f) {
	c_static = f;
    }

    public float getCStatic() {
	return c_static;
    }

    public void setCDynamic(GMathForm m) {
	c_dynamic = new GMathForm(m);
    }

    public GMathForm getCDynamic() {
	return new GMathForm(c_dynamic);
    }

    /** Add a new (pre)condition to the action specification */
    public void addPrecond(int pre, int index) {
	preconds[index] = pre;
    }

    public int numPrecond() {
	return nPre;
    }

    /** Get a (pre)condition given by its index in the list of all (pre)conditions */
    public int getPrecond(int index) {
	return preconds[index];
    }

    /** Get the index of a given (pre)condition. Mostly used for checking if a given
     *  predicate is a (pre)condition of a given action or not */
    public int indexPrecond(int pre) {
	for(int i = 0; i < nPre; i++)
	    if( pre == preconds[i] )
		return i;

	return -1;
    }

    /** For durations of (pre)conditions 
     * Complies with PDDL2.1 level 3: 0 = Instance pre need to be true at start time; 
     * 1 = whole true for the whole duration of action (st,et) */
    public void putPreTime(int time, int index) {
	preTime[index] = time;
    }

    public float getPreTime(int index) {
	return preTime[index];
    }

    /** Add new Test, which is handled similar to predicate precondition */
    public void putTest(GTest aTest, int index) {
	testList[index] = aTest;
    }

    public int numTest() {
	return nTest;
    }

    public GTest getTest(int index) {
	return testList[index];
    }

    public void putTestTime(int time, int index) {
	testTime[index] = time;
    }

    public int getTestTime(int index) {
	return testTime[index];
    }

    /** Add new positive predicate effect */
    public void putAdd(int pred, int index) {
	addList[index] = pred;
    }

    public int numAdd() {
	return nAdd;
    }

    public int getAdd(int index) {
	return addList[index];
    }

    public int indexAdd(int pred) {
	for(int i = 0; i < nAdd; i++)
	    if( addList[i] == pred )
		return i;

	return -1;
    }

    public void putDelete(int pred, int index) {
	deleteList[index] = pred;
    }

    public int numDelete() {
	return nDel;
    }

    public int getDelete(int index) {
	return deleteList[index];
    }

    public int indexDelete(int pred) {
	for(int i = 0; i < nDel; i++)
	    if( deleteList[i] == pred )
		return i;

	return -1;
    }

    /** Specify the time moment in which an effect occur
     *  0: start time; 1: end time; */
    public void putAddTimeEffect(int time, int index) {
	addTime[index] = time;
    }

    public void putDeleteTimeEffect(int time, int index) {
	deleteTime[index] = time;
    }

    public void putSetTimeEffect(int time, int index) {
	setTime[index] = time;
    }

    public int getAddTimeEffect(int index) {
	return addTime[index];
    }

    public int getDeleteTimeEffect(int index) {
	return deleteTime[index];
    }

    public int getSetTimeEffect(int index) {
	return setTime[index];
    }


    /** Add a new "Set", which is a effect that change the value of some
     * continuous function */
    public void putSet(GMySet aSet, int index) {
	setList[index] = aSet;
    }

    public int numSet() {
	return nSet;
    }

    public GMySet getSet(int index) {
	return setList[index];
    }

    /** Function used to shortly print the information about this ground action */
    public String toString() {
	String s = new String();
	Object obj;
	int i;

	s += "(ID: " + getID() + "  ";

	s += ":duration ";
	if( d_type )
	    s += getDStatic();
	else
	    s += getDDynamic();
	s += "\n";

	return s;
    }
}
