/*******************************************************************
    Author: Minh B. Do (Arizona State Univ. - binhminh@asu.edu)
********************************************************************/
package myplanner.basic_ds;
import myplanner.complex_ds.*;

/**
 *  This class can be used for the tree structure that
 *  encode the math formula used in Test/Set/Dynamic Duration/Cost.
 *  the basic math formula can be a float value, or a function value.
 *  The complex structure contain one operator (e.g. +,-,*,/) and
 *  the "left" and "right" elements, which each in turn is a GMathForm.
 *  Similar to MathForm but specialized for Grounded Functions
 */
public class GMathForm {
    int type; // 0: function, 1: value, 2: time (#t), 3: Non Primititive
    float value;  // If type=1.
    int element; // If type=0. This one is the GroundFunction ID

    char operator; // +, -, *, /

    GMathForm left;
    GMathForm right;

    public GMathForm() {
	type = 0;
    }

    public GMathForm(GMathForm g) {
	type = g.getType();
	value = g.getValue();
	element = g.getElement();
	operator = g.getOperator();

	if( type == 3) {
	    left = new GMathForm(g.getLeft());
	    right = new GMathForm(g.getRight());
	}
    }


    /** Set the "Type" of this math formula. 0: function, 1: float value,
     * 2: #t or intermediate time point (not used for level 3, PDDL21),
     * 3: Non primitive or complex math formula */
    public void setType(int b) {
	type = b;
    }

    public int getType() {
	return type;
    }

    /** Set the static value of this GMathForm if the type is type = 1 */
    public void setValue(float f) {
	value = f;
    }

    public float getValue() {
	return value;
    }

    /** Set the "element" of this GMathForm to the Function's ID if the
     *  type of this GMathForm is 0 */
    public void setElement(int e) {
	element = e;
    }

    public int getElement() {
	return element;
    }

    /** Set the operator (+,-,*,/) if this GMathForm is non-primitive */
    public void setOperator(char c) {
	operator = c;
    }

    public char getOperator() {
	return operator;
    }

    public void setLeft(GMathForm m) {
	left = new GMathForm(m);
    }

    public GMathForm getLeft() {
	return new GMathForm(left);
    }

    public void setRight(GMathForm m) {
	right = new GMathForm(m);
    }

    public GMathForm getRight() {
	return new GMathForm(right);
    }

    /** Function to evaluate and return the value of this 
     *  math formula given the current values of all continuous
     *  functions specified in the passed in GMResDB */
    public float value(GMResDB mr, float time) { // #t = time
	// If it's grounded function, need to look at DB 
	if( type == 0 ) {
	    // Need to look up DB for latest updated value
	    return mr.getValue(element);
	}

	// Just a normal Primitive value
	if( type == 1 ) {
	    return value;
	}

	// If it's time (spended from the start-point of an action)
	if( type == 2 ) {
	    // Need to look at time-control module later
	    return time;
	}

	switch (operator) {
	case '+':
	    return left.value(mr,time) + right.value(mr,time);
	case '-':
	    return left.value(mr,time) - right.value(mr,time);
	case '*':
	    return left.value(mr,time) * right.value(mr,time);
	case '/':
	    return left.value(mr,time)/right.value(mr,time);
	default:
	    return 0;
	}
    }

    /** Function to print this GMathForm. Mostly for debugging */
    public String toString() {
	if( type == 0 )
	    return ("funcID(" + element + ")");

	if( type == 1 )
	    return (new Float(value)).toString();

	if( type == 2 )
	    return "#t";

	return "(" +  operator + " " + left.toString() 
	    + " " + right.toString() + ")";
    }
}
