/********************************************************************
   Author: Minh B. Do (Arizona State University - binhminh@asu.edu)
*********************************************************************/
package myplanner.complex_ds;
import myplanner.basic_ds.*;

import java.io.*;
import java.util.*;

/**
 *  GMResDB: Metric Resource Database. This class manage the set of values
 *  for all continuous functions.
 */
public class GMResDB {
    int numFunc;
    float[] funcValues; // Map each FuncID with that function's value

    public GMResDB(int nFunc) {
	numFunc = nFunc;
	funcValues = new float[numFunc];
    }

    public GMResDB(GMResDB mr) {
	numFunc = mr.numFunction();

	funcValues = new float[numFunc];
	for(int i = 0; i < numFunc; i++)
	    funcValues[i] = mr.getValue(i);
    }

    public int numFunction() {
	return numFunc;
    }

    /** Set value for a particular function */
    public void setValue(int funcID, float value) {
	funcValues[funcID] = value;
    }

    /** Get the value of a particula function */
    public float getValue(int funcID) {
	return funcValues[funcID];
    }

    /** Update a function value according to the operator and
     *  the indicated assignment/increment/decreasement value
     *  oper = 0 (=); 1: (-=); 2: (+=); 3: (*=): 4: (/=) */
    public void updateFuncValue(int funcID, float value, int oper) {
	float v = funcValues[funcID];

	if( oper == 0)
	    v = value;

	if( oper == 1)
	    v = v - value;

	if( oper == 2 )
	    v = v + value;

	if( oper == 3)
	    v = v*value;

	if( oper == 4)
	    v = (float) v/value;

	funcValues[funcID] = v;
    }

    /** Function to update the GMResDB if we apply some action. There
     *  are two choices here. One is *instance* access if we want to
     *  apply all changes *instantly* at the beginning of that action
     *  or we want to update it *continuously* during the course of action.
     *  For the PDDL21 Level 3, only the first option is used. */
    public void update(GAction a, boolean instant) {
	GMySet ms;
	int fID;
	float value;

	float dur;

	if( a.numSet() <= 0 )
	    return;

	if( a.getDType() )
	    dur = a.getDStatic();
	else
	    dur = a.getDDynamic().value(this, 0);

	for(int i = 0; i < a.numSet(); i++) {
	    ms = (GMySet) a.getSet(i);
	    fID = ms.getLeftSide();
	    value = getMFValue(ms.getRightSide(), dur);

	    updateFuncValue(fID, value, ms.getAssign());
	}
    }

    /** Function to only update the resource values if the change is increase
     *  used in the heuristics function when we relax the reduction effects
     *  (similar to the delete effects of the predicate) */
    public void relaxUpdate(GAction a) {
	GMySet ms;
	int fID;
	float value;
	int oper;

	float dur;

	if( a.numSet() <= 0 )
	    return;

	if( a.getDType() )
	    dur = a.getDStatic();
	else
	    dur = a.getDDynamic().value(this, 0);

	for(int i = 0; i < a.numSet(); i++) {
	    ms = (GMySet) a.getSet(i);

	    // Don't care if we are reducing the resource level
	    oper = ms.getAssign();
	    if( (oper == 1) )
		continue;
	    
	    fID = ms.getLeftSide();
	    value = getMFValue(ms.getRightSide(), dur);

	    float v = funcValues[fID];

	    if( oper == 0 && v < value) {
		funcValues[fID] = value;
	    } else if( oper == 2 ) {
		funcValues[fID] = v+value;
	    } else  if( oper == 3 && value > 1) {
		funcValues[fID] = v*value;
		v = v*value;
	    } else if( oper == 4 &&  value < 1) {
		funcValues[fID] = v/value;
	    }
	}
	
    }

    /** Function to return a value of a MathForm given the current
     *  values of all functions in the GMResDB.  */
    public float getMFValue(GMathForm mf, float time) {
	return mf.value(this, time); // Test, will NOT involve #t in test
    }

    /** Function to check if the metric-precond (tests) of an action are
     *  satisfied given the current state of the GMResDB. */
    public boolean applicable(GAction action) {
	if( action.numTest() == 0 )
	    return true;

	GTest test;
	float left, right;
	for(int i = 0; i < action.numTest(); i++) {
	    test = action.getTest(i);
	    left = getValue(test.getLeftSide());
	    right = getMFValue(test.getRightSide(), 0);

	    if( test.getComparator() == 0 && left != right )
		return false;

	    if( test.getComparator() == 1 && left >= right)
		return false;

	    if( test.getComparator() == 2 && left > right)
		return false;

	    if( test.getComparator() == 3 && left <= right)
		return false;

	    if( test.getComparator() == 4 && left < right)
		return false;

	}
	return true;
    }
}
