/************************************************************
   Author: Daniel Bryce -- Arizona State University
*************************************************************/
package myplanner.gui;

import myplanner.gui.*;
import javax.swing.*;
import java.awt.*;
import java.awt.Graphics.*;
import java.awt.Graphics2D.*;
import java.math.*;
import java.awt.geom.*;
import java.util.*;
import java.awt.event.*;
import myplanner.utils.OrderRelation;
import myplanner.utils.Ordering;
import myplanner.utils.GreedyPost;
import myplanner.parsing.Grounding;



public class GanttChart extends JPanel implements MouseListener{

    //Display options
    boolean showCausalLinks;
    boolean showPredMutex;
    boolean showResMutex;
    boolean showUnitActions;
    boolean removeRedundantOrderings;
    
    public boolean ganttHighlight = false;
    
    //Sibling
    ActionWindow actionWindow;
    
    //Plan Info
    int highlight;
    public Vector pre_highlight;
    public Vector post_highlight;
    public Vector pre_highlight_rel;
    public Vector post_highlight_rel;
    
    int plansize;
    Grounding grounding;
    GreedyPost greedyPost;
    ArrayList act_cells;
    Ordering[] order;
    Ordering[] orderCopy;
    ArrayList actSigList;
    ArrayList actions;
    float[] times;
    float[] durs;

    double y_max_step_size = 100.0;


    int x_origin;
    int y_origin;
    int x_boundary;
    int y_boundary;
    int x_timeline_origin;
    int y_timeline_origin;
    int x_timeline_boundary;
    int y_timeline_boundary;
    int x_diagram_origin;
    int y_diagram_origin;
    int y_timeline_diagram_padding;
    int x_diagram_boundary;
    int y_diagram_boundary;
    
    double y_step_width;
    


    public GanttChart(int width, int height, GreedyPost gp, Grounding gi, boolean[] options) {
	super(true);
	actionWindow = null;
        grounding = gi;
        greedyPost = gp;
        actions = greedyPost.aPlan;
        actSigList = greedyPost.actSigList;
	order = greedyPost.orderingSet;
	times = greedyPost.startTimes;
	durs = greedyPost.actDurs;
	plansize = greedyPost.planSize;
	highlight = -1;

	act_cells = new ArrayList();
	JPanel tmp;
	for (int i = 0; i < actions.size(); i++) {
	    tmp = new JPanel();
	    tmp.addMouseListener(this);
	    act_cells.add(tmp);
	}
	
	showCausalLinks = options[0];
	showPredMutex = options[1];
	showResMutex = options[2];
	showUnitActions = options[3];
	removeRedundantOrderings = options[4];
	
	pre_highlight = new Vector();
	post_highlight = new Vector();
	pre_highlight_rel = new Vector();
	post_highlight_rel = new Vector();
	
	x_origin = 0;
	y_origin = 0;
	x_boundary = width;
	y_boundary = height;
	x_timeline_origin = 0;
	y_timeline_origin = 0;
	x_timeline_boundary = x_boundary;
	y_timeline_boundary = 50;
	x_diagram_origin = 0;
	y_diagram_origin = y_timeline_boundary + y_timeline_diagram_padding;
	y_timeline_diagram_padding = 0;
	x_diagram_boundary = x_boundary;
	y_diagram_boundary = y_boundary;

	y_step_width = Math.min((y_diagram_boundary - y_diagram_origin)/(plansize + 1) ,
				y_max_step_size);


	setBounds(x_origin, y_origin, x_boundary, y_boundary);
	setBackground(java.awt.Color.white);

	setVisible(true);
    }

    public int getStepHeight() {
	return (int)y_step_width;
    }

    public void setDrawingOptions(boolean[] options){
	showCausalLinks = options[0];
	showPredMutex = options[1];
	showResMutex = options[2];
	showUnitActions = options[3];
	removeRedundantOrderings = options[4];
	
	int p = highlight;

	if(p > 0) {
	    highlightCell(highlight);
	    highlightCell(p);
	}
	repaint();
    }

    public void newSize(int width) {
	newSize(width, y_boundary);
    }

    
    public void newSize(int width, int height) {
	x_boundary = width;
	y_boundary = height;
	x_timeline_boundary = x_boundary;
	x_diagram_boundary = x_boundary;
	y_diagram_boundary = y_boundary;
	setBounds(x_origin, y_origin, x_boundary, y_boundary);
	setPreferredSize(new Dimension(width, height));
    }
    
    public int getWidth() { return x_boundary;}
    public int getHeight() { return y_boundary;}
    
    public void paint (Graphics gantt) {
	Graphics2D chart = (Graphics2D) gantt;
	super.paint(chart);
		
	
	////////////////gantt cell construction
	int i = 0;
	double x_cell_size =0.0;
	double y_cell_size;
	double x_cell_origin;
	double y_cell_origin;

	double x_min_step_width = 10.0;

	double x_step_padding = 9.0;
	double x_step_width;
	y_step_width = Math.min((y_diagram_boundary - y_diagram_origin)/(plansize + 1) ,
				y_max_step_size);
	double y_step_padding = 0.2 * y_step_width;
	double x_step_origin = x_diagram_origin;
	double y_step_origin = y_diagram_origin;

	int x_scale = (int)(1000*(x_diagram_boundary - x_diagram_origin)/
			    (roundUp(1000*endTime())));


	if(!showUnitActions) {
	    
	    //timeline ops
	    chart.setPaint(Color.lightGray);
	    chart.fill3DRect(x_timeline_origin,
			     y_timeline_origin,
			     x_timeline_boundary - x_timeline_origin,
			     y_timeline_boundary - y_timeline_origin,
			     true);
	    //System.out.println(x_timeline_origin + " " + y_timeline_origin);
	    double half_timeline = (y_timeline_boundary + y_timeline_origin)/2;
	    chart.setPaint(Color.black);
	    chart.draw(new Line2D.Double(x_timeline_origin, 
					 half_timeline, 
					 x_timeline_boundary, 
					 half_timeline));
	    
	    double skip = roundUp(Math.sqrt(endTime()));
	    double scale_marker = x_scale*skip;
	    double minor_scale_marker = 0.0;
	    while(scale_marker < x_timeline_boundary) {
		chart.draw(new Line2D.Double(scale_marker, 
					     y_timeline_origin, 
					     scale_marker, 
					     y_timeline_boundary));
		chart.drawString(String.valueOf((int)(scale_marker/x_scale)), 
				 (int)scale_marker, 
				 (int)(0.75*half_timeline));
		scale_marker += x_scale*skip;
		while (minor_scale_marker < scale_marker) {
		    chart.draw(new Line2D.Double(minor_scale_marker, 
						 0.75*half_timeline, 
						 minor_scale_marker, 
						 1.25*half_timeline));
		    minor_scale_marker += x_scale;
		}
	    }
	}

	//draw the actions
	for (i = 0; i < actions.size(); i++) {
	    
	    if(!showUnitActions) {
		x_step_width = Math.max(durs[i] * x_scale,
					x_min_step_width);
		x_step_origin = (times[i] * x_scale);
	    }
	    else {
		x_step_width = (x_boundary - x_origin)/plansize;
		x_step_origin = (i) * x_step_width;
	    }
	    
	    x_cell_size = x_step_width - x_step_padding;
	    y_cell_size = y_step_width - y_step_padding;
	    x_cell_origin = x_step_origin + 0.5*x_step_padding;
	    y_cell_origin = y_step_origin + 0.5*y_step_padding;
	    
	    //determine what color to draw action
	    if (highlight == i)
		chart.setPaint(Color.red);
	    else if(pre_highlight.contains(new Integer(i)))
		chart.setPaint(Color.green);
	    else if(post_highlight.contains(new Integer(i)))
		chart.setPaint(Color.pink);
	    else
		chart.setPaint(Color.yellow);
	    
	    ((JPanel)act_cells.get(i)).setLocation((int)x_cell_origin,
						     (int)y_cell_origin);
	    ((JPanel)act_cells.get(i)).setSize((int)x_cell_size, (int)y_cell_size);
	    chart.fill(new Rectangle2D.Double(x_cell_origin,
					      y_cell_origin,
					      x_cell_size ,
					      y_cell_size));
	    chart.setPaint(Color.black);
	    chart.draw(new Rectangle2D.Double(x_cell_origin,
					      y_cell_origin,
					      x_cell_size ,
					      y_cell_size));
	    
	    String tmp_act_out = ((String)actSigList.get(i));
	    tmp_act_out = tmp_act_out.substring(0, tmp_act_out.indexOf("*"));
	    
	    chart.drawString(tmp_act_out, 
			     (int)x_cell_origin, 
			     (int)(y_cell_origin + y_cell_size));
	    chart.setPaint(Color.gray);
	    
	    if(!showUnitActions)
		chart.draw(new Line2D.Double(x_step_origin, y_origin, x_step_origin, y_diagram_boundary));
	    
	    chart.draw(new Rectangle2D.Double(x_cell_origin,
					      y_cell_origin,
					      x_cell_size ,
					      y_cell_size));
	
	    this.add(((JPanel)act_cells.get(i)));

	    //update y coord
	    y_step_origin += y_step_width;
	}

	int x_arc_start;
	int y_arc_start;
	int x_arc_end;
	int y_arc_end;
	Color col = Color.black;
	int arc_padding = 0;

	//Process Ordering relations
	for (i = 0; i < plansize ; i++) {
	    //chart.draw(new Line2D.Double(x_step_origin, y_origin, x_step_origin, y_diagram_boundary));
	    //System.out.println("Step: " + i + " #of order constraints: " + order[i].relList.size());
	    arc_padding = 0;
	    for (int j = 0; j < order[i].relList.size(); j++) {
		//System.out.println("Order with: " + ((OrderRelation)order[i].relList.get(j)).actIndex );
		if((!removeRedundantOrderings  || 
		    !order[i].redOrder.contains(new Integer(j)))){
		    
		    if(((OrderRelation)order[i].relList.get(j)).actIndex > 0){
			
			x_arc_start = ((JPanel)act_cells.get(((OrderRelation)order[i].relList.get(j)).actIndex)).getX();
			
			if( ((OrderRelation)order[i].relList.get(j)).saTime == 1)
			
			    x_arc_start += ((JPanel)act_cells.get(((OrderRelation)order[i].relList.get(j)).actIndex)).getWidth();
			x_arc_end = ((JPanel)act_cells.get(i)).getX();
		    }

		    else {

			x_arc_start = (int)(x_step_padding*0.5);
			if( ((OrderRelation)order[i].relList.get(j)).saTime == 1)
			    x_arc_start += x_cell_size;
			x_arc_end = ((JPanel)act_cells.get(i)).getX();
		    }

		    if( ((OrderRelation)order[i].relList.get(j)).eaTime == 1)
			x_arc_end += ((JPanel)act_cells.get(i)).getWidth();

		    y_arc_start = ((JPanel)act_cells.get(((OrderRelation)order[i].relList.get(j)).actIndex)).getY() + 
			(int)(0.5*y_step_width);
		    y_arc_end = ((JPanel)act_cells.get(i)).getY();

		    if (highlight < 0 || 
			i == highlight || 
			(post_highlight.contains(new Integer(i))) && 
			 ((OrderRelation)order[i].relList.get(j)).actIndex == highlight) {
			switch(((OrderRelation)order[i].relList.get(j)).relation) {

			case(1):	{
			    if(!showCausalLinks) col = Color.gray;
			    else col = Color.red;
			    break;
			}
			case(2):	{
			    if(!showPredMutex)	col =  Color.gray;
			    else col = Color.blue;
			    break;
			}
			case(3):	{
			    if(!showResMutex) col =  Color.gray;
			    else col = Color.green;
			    break;
			}

			}

		    }

		    else
			col = Color.gray;
		    
		    chart.setPaint(col);
		    
		    //Draw links
		    if(col != Color.gray) {
			if(((OrderRelation)order[i].relList.get(j)).saTime == 1) {
			    chart.draw(new Line2D.Double(x_arc_start,
							 y_arc_start + arc_padding,
							 x_arc_start + 0.25*x_step_padding + arc_padding,
							 y_arc_start + arc_padding));
			    
				    
			    chart.draw(new Line2D.Double(x_arc_start + 0.25*x_step_padding + arc_padding,
							 y_arc_start + arc_padding,
      							 x_arc_end,
							 y_arc_start + arc_padding));

			    chart.draw(new Line2D.Double(x_arc_end,
							 y_arc_start + arc_padding,
							 x_arc_end ,
							 y_arc_end - 0.5 * y_step_padding + arc_padding));

			    chart.draw(new Line2D.Double(x_arc_end ,
							 y_arc_end - 0.5 * y_step_padding + arc_padding,
							 x_arc_end ,
							 y_arc_end));
			}
			else {
			    chart.draw(new Line2D.Double(x_arc_start,
							 y_arc_start + arc_padding,
							 x_arc_start - 0.25*x_step_padding - arc_padding,
							 y_arc_start + arc_padding));
			    chart.draw(new Line2D.Double(x_arc_start - 0.25*x_step_padding - arc_padding,
							 y_arc_start + arc_padding,
							 x_arc_start - 0.25*x_step_padding - arc_padding,
							 y_arc_end - 0.5 * y_step_padding + arc_padding));
			    chart.draw(new Line2D.Double(x_arc_start - 0.25*x_step_padding - arc_padding,
							 y_arc_end - 0.5 * y_step_padding + arc_padding,
							 x_arc_end ,
							 y_arc_end - 0.5 * y_step_padding + arc_padding));
			    chart.draw(new Line2D.Double(x_arc_end ,
							 y_arc_end - 0.5 * y_step_padding + arc_padding,
							 x_arc_end ,
							 y_arc_end));
			}
		    }
		    arc_padding = 3;
		}
	    }
	}
    }

    private double roundUp(double num) {
	int n = (int)num*1000;
	while((n/1000)%5 != 0)
	    n++;
	
	return n/1000;	
    }

    private double endTime() {
	double end = 0.0;
	double x = 0.0;
	for (int i = 0; i < actions.size(); i++) {
	    x = (times[i] + durs[i]);
	    if (x > end)
		end = x;
	}
	return end;
	
    }

    public void highlightCell(int i) {

	if(i > -1) {
	    //System.out.println("Highlight: " + i);
	    pre_highlight.removeAllElements();
	    post_highlight.removeAllElements();
	    pre_highlight_rel.removeAllElements();
	    post_highlight_rel.removeAllElements();
	    
	    if (highlight == i) {
		highlight = -1;
		actionWindow.unhighlight();
		ganttHighlight = false;
	    }

	    else{
		
		ganttHighlight = true;
		highlight = i;

		//get pre_highlight
		int tmp;
		int rel;
		for (int j = 0; j < order[highlight].relList.size(); j++) {
		    tmp = ((OrderRelation)order[highlight].relList.get(j)).actIndex;
		    rel = ((OrderRelation)order[highlight].relList.get(j)).relation;
		    if (!removeRedundantOrderings  || !order[highlight].redOrder.contains(new Integer(j))){
			pre_highlight.add(new Integer(tmp));
			pre_highlight_rel.add(new String(getRelString(rel)));
		    }
		}

		//get post_highlight;
		for (int k = 0; k < order.length; k++) {
		    for (int j = 0; j < order[k].relList.size(); j++){
			tmp = ((OrderRelation)order[k].relList.get(j)).actIndex;
			rel = ((OrderRelation)order[k].relList.get(j)).relation;
			if (tmp == highlight &&	
			    (!removeRedundantOrderings  || 
			     !order[k].redOrder.contains(new Integer(j)))) {
			    post_highlight.add(new Integer(k));
			    post_highlight_rel.add(getRelString(rel));
			}
		    }
		}
		actionWindow.setActionDetailPanel(i);
	    }
	}
	this.repaint();
    }
    



    String getRelString(int i) {
	switch(i) {
	case(1):	{
	    return "<font color=#FF0000 >Causal Link</font>";
	}
	case(2):	{
	    return "<font color=#0000FF >Predicate Mutex</font>";
	}
	case(3):	{
	    return "<font color=#00FF00>Resource Mutex</font>";
	}
	}
	return "NADA";
    }
    
    public void removeRedundant() {
    }
    
    public void setAW(ActionWindow aw) {
	actionWindow = aw;
    }

    //Mouse Listener
    public void mouseClicked(MouseEvent e) {
	Component c = e.getComponent();
	for (int i = 0; i < act_cells.size(); i++) {
	    if (c == ((JPanel)act_cells.get(i))) {
		//System.out.println("highlight gc");
		highlightCell(i);
		
	    }
	}
    }
    public void mousePressed(MouseEvent e) {}
    public void mouseReleased(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    //Mouse Listener

}
